#!/usr/bin/env python
import sliderepl
sliderepl.SADeck.run()

### slide:: 1 Setup

from sqlalchemy import *

meta = MetaData(create_engine('sqlite://'))

from sqlalchemy.orm import *
Session = scoped_session(sessionmaker(transactional=True,autoflush=False))

class Base(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.iteritems():
            setattr(self, key, value)
    def __repr__(self):
        cls = self.__class__
        return "%s(%s)" % (
            cls.__name__,
            ", ".join(["%s=%r" % (c, getattr(self, c)) for c in cls.c.keys()])
        )
    query = Session.query_property()

### slide:: 2 Table Setup

companies = Table('companies', meta,
   Column('company_id', Integer, primary_key=True),
   Column('name', String(50)))

people = Table('people', meta,
   Column('person_id', Integer, primary_key=True),
   Column('company_id', Integer, ForeignKey('companies.company_id')),
   Column('name', String(50)),
   Column('type', String(30)))

engineers = Table('engineers', meta,
   Column('person_id', Integer, ForeignKey('people.person_id'), primary_key=True),
   Column('primary_language', String(50)),
  )

managers = Table('managers', meta,
  Column('person_id', Integer, ForeignKey('people.person_id'), primary_key=True),
  Column('golf_swing', String(30)),
 )

machines = Table('machines', meta,
    Column('machine_id', Integer, primary_key=True),
    Column('name', String(50)),
    Column('engineer_id', Integer, ForeignKey('engineers.person_id')))


meta.create_all()

### slide:: 3 Classes

class Company(Base):
    pass

class Person(Base):
    pass

class Engineer(Person):
    pass

class Manager(Person):
    pass

class Machine(Base):
    pass

### slide:: 4 Mappers

mapper(Company, companies, properties={
    'employees':relation(Person)
})
mapper(Person, people, polymorphic_on=people.c.type)
mapper(Engineer, engineers, inherits=Person, polymorphic_identity='engineer', properties={
    'machines':relation(Machine)
})
mapper(Manager, managers, inherits=Person, polymorphic_identity='manager')
mapper(Machine, machines)

### slide:: 5 Data

c1 = Company(name="MegaCorp, Inc.", employees=[
    Engineer(name="dilbert", primary_language="java", 
            machines=[
                Machine(name='IBM ThinkPad'),
                Machine(name='IPhone'),
            ]),

    Engineer(name="wally", engineer_name="wally",
            primary_language="c++", 
            machines=[
                Machine(name="Commodore 64")
            ]),

    Manager(name="dogbert", manager_name="dogbert", golf_swing="fore!")
])

c2 = Company(name="Elbonia, Inc.", employees=[
    Engineer(name="vlad", primary_language="cobol", 
            machines=[
                Machine(name="Commodore 64"),
                Machine(name="IBM 3270")
        ])
])

Session.save(c1)
Session.save(c2)
Session.flush()

### slide:: 6 Polymorphic Querying

Session.clear()

print Person.query.filter(Person.name=='dilbert').all()

### slide:: 7

Session.clear()

print Person.query.with_polymorphic('*').filter(Person.name=='dilbert').all()

### slide:: 8

# issuing a query against Person returns Engineer
# and Manager objects (and Person, if those exist)
print Person.query.all()

### slide:: 9

# To get just Engineers (and subclasses of Engineer), # query from Engineer
print Engineer.query.all()

### slide:: 10

# Filter criterion can be against the classes explicitly present:
print Person.query.filter(Person.name=='dilbert').all()

### slide:: 11

print Engineer.query.filter(Person.name=='dilbert').\
    filter(Engineer.primary_language=='Java').all()

### slide:: 12

print Person.query.with_polymorphic(Engineer).\
    filter(Engineer.primary_language=='Java').all()

### slide:: 13 Polymorphic Relations

# suppose we want to load Company objects,
# which contain a certain person.  The any()
# operator creates an EXISTS clause:
print Company.query.filter(
   Company.employees.any(Person.name=='dilbert')
 ).all()


### slide:: 14

# what if we want to filter on a subclass criterion ?
# use of_type().  the EXISTS will join against the
# subclass' table as well.
print Company.query.filter(
   Company.employees.of_type(Manager).\
     any(Manager.golf_swing=="fore!")
 ).all()

### slide:: 15 Polymorphic Joins

# similar rules apply when we deal with joins.
# if we want to join Companies to Employees,
# we can use join():
print Company.query.join('employees').\
        filter(Person.name=='dilbert').all()

### slide:: 16

# to join to Engineers via the "employees"
# relation, again the easiest way is of_type():
print Company.query.\
        join(Company.employees.of_type(Engineer)).\
        filter(Engineer.primary_language=="c++").all()

### slide:: 17

# same idea works when joining further:
print Company.query.\
        join([
            Company.employees.of_type(Engineer),
            'machines'
        ]).filter(Machine.name=='Commodore 64').all()

### slide:: 18 Exercise Answers

print Engineer.query.filter(Engineer.machines.any(Machine.name=='IBM 3270')).all()

### slide:: 19

print Company.query.join(Company.employees.of_type(Engineer)).filter(Engineer.machines.any(Machine.name=='IBM 3270')).all()

### slide:: 20

print Company.query.filter(~Company.employees.of_type(Manager).any()).all()

### slide:: 21

print Person.query.with_polymorphic('*').filter(or_(Engineer.primary_language=='c++', Manager.type=='manager')).all()

### slide:: end
