#!/usr/bin/env python
import sliderepl
sliderepl.SADeck.run()

### slide:: 1

from sqlalchemy import *
from sqlalchemy.orm import *
from decimal import Decimal
from datetime import date

metadata = MetaData('sqlite://')
session = create_session()

def fixture(table, columns, rows):
    def onload(event, schema_item, connection):
        insert = table.insert()
        column_names = [col.key for col in columns]
        connection.execute(insert, [dict(zip(column_names, column_values))
                                    for column_values in rows])
    table.append_ddl_listener('after-create', onload)

def reset():
    clear_mappers()
    global session
    session = create_session()

customers = Table('customers', metadata,
                  Column('id', Integer, primary_key=True),
                  Column('name', String(64)))
fixture(customers, customers.c,
        [(1, 'custie 1'), (2, 'custie 2')])

orders = Table('orders', metadata,
               Column('id', Integer, primary_key=True),
               Column('customer_id', Integer,
                      ForeignKey('customers.id')))
fixture(orders, orders.c,
        [(1, 1), (2, 1), (3, 2)])

items = Table('items', metadata,
              Column('id', Integer, primary_key=True),
              Column('name', String(64)),
              Column('retail_price', Numeric(10, 2)))
fixture(items, items.c,
        [(1, 'plastic foo', Decimal('10.00')),
         (2, 'iron bar', Decimal('24.99')),
         (3, 'bucky baz', Decimal('49.50'))])

pricing = Table('pricing', metadata,
                Column('item_id', Integer, ForeignKey('items.id'),
                       primary_key=True, autoincrement=False),
                Column('begin', Date, primary_key=True),
                Column('end', Date, primary_key=True),
                Column('price', Numeric(10, 2), nullable=False))
fixture(pricing, pricing.c,
        [(1, date(2007, 1, 1), date(2009, 1, 1), Decimal('10.00')),
         (2, date(2007, 1, 1), date(2008, 1, 1), Decimal('19.99')),
         (2, date(2008, 1, 1), date(2009, 1, 1), Decimal('24.99')),
         (3, date(2007, 1, 1), date(2008, 3, 15), Decimal('49.50')),
         (3, date(2008, 3, 15), date(2009, 1, 1), Decimal('59.50'))])

orders_to_items = Table('orders_to_items', metadata,
                        Column('order_id', Integer, ForeignKey('orders.id'),
                               primary_key=True),
                        Column('item_id', Integer, ForeignKey('items.id'),
                               primary_key=True))
fixture(orders_to_items, orders_to_items.c,
        [(1, 1), (2, 1), (2, 2), (3, 1), (3, 2), (3, 3)])

order_items = Table('order_items', metadata,
                    Column('order_id', Integer, ForeignKey('orders.id'),
                           primary_key=True),
                    Column('item_id', Integer, ForeignKey('items.id'),
                           primary_key=True),
                    Column('purchase_price', Numeric(10, 2), nullable=False))
fixture(order_items, order_items.c,
        [(1, 1, 1, Decimal('10.00')),
         (2, 1, 1, Decimal('10.00')),
         (2, 2, 2, Decimal('19.99')),
         (3, 1, 1, Decimal('10.00')),
         (3, 2, 2, Decimal('24.99')),
         (3, 3, 3, Decimal('49.50'))])

class Base(object):
    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)
class Customer(Base):
    def __repr__(self):
        return '<Customer #%r %r>' % (self.id, self.name)
class Order(Base):
    def __repr__(self):
        return '<Order #%r of %r>' % (
            self.id, getattr(self, 'customer', self.customer_id))
class Item(Base):
    def __repr__(self):
        return '<Item #%r %r>' % (self.id, self.name)
class OrderItem(Base):
    def __repr__(self):
        return '<OrderItem #%r-#%r @ %s>' % (self.order_id, self.item_id,
                                             self.purchase_price)
class Pricing(Base):
    def __repr__(self):
        return '<Pricing %s from %s to %s>' % (
            self.price, self.begin, self.end)

metadata.create_all()

### slide:: 2

mapper(Customer, customers, properties={
        'orders': relation(Order, backref='customer')})
mapper(Order, orders)

### slide:: 3

c = Customer(name='custie')
o = Order()

c.orders.append(o)
c.orders

session.save(c)
o in session

### slide:: 4

reset()

class MyCollection(list):
    pass

mapper(Customer, customers, properties={
        'orders': relation(Order, backref='customer',
                           cascade='all,delete-orphan',
                           collection_class=MyCollection)})
mapper(Order, orders)

c = Customer(name='custie')
# type(c.orders)

### slide:: 5

o = Order()
c.orders.append(o)

session.save(c)
o in session

session.flush()
c.orders.pop()
session.flush()

### slide:: 6

reset()

mapper(Order, orders)
mapper(Customer, customers, properties={
        'orders': relation(Order,
                           backref='customer',
                           lazy=False,
                           collection_class=set)})

# hint: to load some Customers, try session.query(Customer).all()

### slide:: 7

reset()

mapper(Customer, customers, properties={
        'orders': relation(Order, backref='customer')})
mapper(Order, orders, properties={
        'items': relation(Item, secondary=orders_to_items,
                          cascade='all, delete-orphan',
                          backref='orders')})
mapper(Item, items)

o = Order()
ia = Item(name='item a')
ib = Item(name='item b')
o.items.extend([ia, ib])

### slide:: 8

reset()
mapper(Order, orders, properties={
        'orderitems': relation(OrderItem, backref='order')
        })
mapper(OrderItem, order_items, properties={
        'item': relation(Item)
        })
mapper(Item, items)

### slide:: 9

order = session.query(Order).first()
order.orderitems

oi = order.orderitems[0]
oi.order is order
oi.item
oi.purchase_price

### slide:: 10

[x.item for x in order.orderitems]

### slide:: 11

reset()

from sqlalchemy.ext.associationproxy import association_proxy
class Order2(Order):
    items = association_proxy('orderitems', 'item')

mapper(Order2, orders, properties={
        'orderitems': relation(OrderItem, backref='order')
        })
mapper(OrderItem, order_items, properties={
        'item': relation(Item, lazy=False)
        })
mapper(Item, items)

order = session.query(Order2).first()
order.orderitems
order.items

### slide:: 12

reset()
from sqlalchemy.orm.collections import collection

class TimeVaryingAttribute(object):
    def __init__(self):
        self.members = []

    def __getitem__(self, when):
        for tv in self.members:
            if tv.begin <= when and tv.end > when:
                return tv.price
        raise IndexError("No value for date %s" % when)

    def today(self):
        return self[date.today()]

    def __setitem__(self, (begin, end), value):
        self._splice(begin, end, value)

    def _splice(self, begin, end, value):
        for tv in self.members:
            if tv.begin < begin and tv.end >= end:
                # Split existing entry into pre- and post- sections.
                self.members.append(
                    Pricing(begin=end, end=tv.end, price=tv.price))
                tv.end = begin
                break
        i = 0
        while i < len(self.members):
            tv = self.members[i]
            # Remove wholly overlapped entries
            if tv.begin >= begin and tv.end <= end:
                self._remove(tv)
                continue
            # Truncate tail overlaps
            if tv.begin < end and tv.end > end:
                tv.begin = end
                continue
            i += 1
        self._append(Pricing(begin=begin, end=end, price=value))
        self.members.sort()
        
    @collection.appender
    def _append(self, member):
        self.members.append(member)
    @collection.remover
    def _remove(self, member):
        self.members.remove(member)
    @collection.iterator
    def _iter(self):
        return iter(self.members)

mapper(Item, items, properties={
        'price_for': relation(Pricing, backref='item',
                              collection_class=TimeVaryingAttribute)
        })
mapper(Pricing, pricing)

### slide:: 13

item = session.query(Item).get(3)
item.price_for.today()
item.price_for[date(2008, 10, 10)]

# Special for October: only 1.00
item.price_for[date(2008, 10, 1), date(2008, 11, 1)] = Decimal('1.00')

item.price_for[date(2008, 10, 10)]
item.price_for[date(2008, 9, 15)]
item.price_for[date(2008, 11, 15)]

### slide:: end
