#!/usr/bin/env python
import sliderepl
sliderepl.SADeck.run()

### slide:: 1 Setup

from sqlalchemy import *

meta = MetaData(create_engine('sqlite://'))

users = Table('users', meta,
    Column('id', Integer, primary_key=True),
    Column('name', String(50))
)

addresses = Table('addresses', meta,
    Column('id', Integer, primary_key=True),
    Column('email', String(50)),
    Column('user_id', Integer, ForeignKey('users.id'))
)

meta.create_all()

### slide:: 2

ins = users.insert()

ins.execute([
   {'name':'jack'},
   {'name':'ed'},
   {'name':'wendy'},
   {'name':'mary'}
])

ins = addresses.insert()
ins.execute([
   {'user_id':1, 'email':'jack@jack.com'},
   {'user_id':2, 'email':'ed@yahoo.com'},
   {'user_id':2, 'email':'ed@msn.com'},
   {'user_id':3, 'email':'wendy@nyt.com'},
])

### slide:: 3 Generative Selects

sel = select([users.c.id])

sel = sel.select_from(
      users.join(addresses))

sel = sel.where(
    or_(users.c.name=='ed',
      addresses.c.email.like('ed%')))

sel = sel.column(addresses.c.email)

sel = sel.order_by(asc(users.c.name))

for row in sel.execute():
   print row


### slide:: 4 Generative Updates, Deletes

users.update().where(users.c.name=='wendy').\
   values({'name':'wendy@nypost.com'}).execute()

del_ = users.delete().where(users.c.name == 'wendy')
print del_

### slide:: 5 Answers

print select([users]).\
    where(users.c.id>5)

print users.update().where(users.c.id==10).\
    values({'name':'jack smith'})

print select([users, addresses]).\
    select_from(users.join(addresses)).\
    order_by(desc(addresses.c.email))

print users.delete().where(
    addresses.c.user_id==
         select([users.c.id]).
         where(users.c.name=='jack')
    )
### slide:: 6 Scalar Subqueries

sel = select([users.c.id]).where(users.c.name=='jack')

print [c for c in sel.c]  # sel has a 'c' collection

### slide:: 7

sel = sel.as_scalar()

try:
    print [c for c in sel.c]   # scalar has no 'c', raises error
except:
    import sys
    print sys.exc_info()[1]

print sel + 3   # but now it can do this

### slide:: 8 Scalar Subqueries

sel = select([users.c.id]).where(users.c.name=='jack').as_scalar()

# column clause subqueries
for row in select([users, sel]).execute():
    print row

### slide:: 9

for row in select([sel, sel + 5, sel + 7], bind=meta.bind).execute():
    print row

### slide:: 10

# WHERE clause subquery
for row in select([users]).where(users.c.id==sel).execute():
    print row

### slide:: 11

for row in select([users]).where(users.c.id==sel + 1).execute():
    print row

### slide:: 12 Row-Based Subqueries

sel = select([users]).where(users.c.name.like('%e%')).alias()

for row in select([users]).select_from(
        users.join(sel, users.c.id==sel.c.id)
    ).execute():
    print row

### slide:: 13 Implicit Correlation

# select the minimum 'email' string for each user
adr = select([func.min(addresses.c.email)]).\
    where(addresses.c.user_id==users.c.id).as_scalar()
print adr

### slide:: 14

# return rows from users, combined with the first email
# address for each
sel = select([users, adr])

for row in sel.execute():
    print row

### slide:: 15 Explicit Correlation

# same subquery, except we say "correlate(users)" so
# that only "users" correlates
adr = select([func.min(addresses.c.email)]).\
    where(addresses.c.user_id==users.c.id).\
    correlate(users).as_scalar()
print adr

### slide:: 16

# select full rows from users and addresses,
# using the first address row for each
sel = select([users, addresses]).\
    select_from(users.outerjoin(addresses)).\
    where(or_(addresses.c.email==None, addresses.c.email==adr))

for row in sel.execute():
    print row

### slide:: 17 Exercise Answers

# 1. Write this SQL as a SQLAlchemy expression.
#
# SELECT id, name FROM users
# WHERE id=(select user_id FROM addresses WHERE email='foo@bar.com')

print select([users]).where(users.c.id==
    select([addresses.c.user_id]).
    where(addresses.c.email=='foo@bar.com').\
    as_scalar()
  )

### slide:: 18

# 2. Given this statement:
#
# SELECT id, user_id, email FROM addresses WHERE email IN ('jack@jack.com', 'ed@msn.com')
#
# Convert the above to a SQLAlchemy expression, then embed 
# it within an expression that joins the "users" table to 
# the query in the FROM clause, and returns user's name and email address.

sel = select([addresses]).where(addresses.c.email.in_(['jack@jack.com', 'ed@msn.com'])).alias()
for row in select([users.c.name, sel.c.email]).select_from(users.join(sel, users.c.id==sel.c.user_id)).execute():
    print row


### slide:: 19 exists() expressions

print select([users]).where(
   exists([1]).
   where(addresses.c.user_id==users.c.id).
   where(addresses.c.email=='foo@bar.com')
)

### slide:: 20

print select([users]).where(
    ~exists([1]).
    where(addresses.c.user_id==users.c.id)
  )
  
### slide:: 21 Exercise Answers

# 3. rewrite 
# SELECT id, name FROM users WHERE EXISTS (SELECT 1 FROM addresses
#  WHERE user_id=users.id AND email IN ('jack@yahoo.com', 'jack@msn.com'))
# 
# SELECT id, name FROM users WHERE EXISTS (SELECT 1 FROM addresses
#  WHERE user_id=users.id AND email='jack@yahoo.com')
#  AND EXISTS (SELECT 1 FROM addresses WHERE user_id=users.id AND
#  email='jack@msn.com')
#
# as SQLAlchemy expressions

print select([users]).where(
    exists([1]).where(addresses.c.user_id==users.c.id).\
    where(or_(addresses.c.email.in_(['jack@yahoo.com', 'jack@msn.com'])))
    )

### slide:: 22

print select([users]).where(
          exists([1]).where(addresses.c.user_id==users.c.id).\
          where(addresses.c.user_id=='jack@yahoo.com')
        ).where(
          exists([1]).where(addresses.c.user_id==users.c.id).\
          where(addresses.c.user_id=='jack@msn.com')
        )


### slide:: end
